/***  Lightness1024! ProgrammatO E.U.R.L.    ***/
/*** Header de la classe template SecureArray ***/
/*** version rvise le 01/11/2005 17:57  :49 ***/

/*
----------------------------------------------------
Ce logiciel est rgi par la licence CeCILL soumise au droit franais et
respectant les principes de diffusion des logiciels libres. Vous pouvez
utiliser, modifier et/ou redistribuer ce programme sous les conditions
de la licence CeCILL telle que diffuse par le CEA, le CNRS et l'INRIA 
sur le site "http://www.cecill.info".
(...)
Le fait que vous puissiez accder  cet en-tte signifie que vous avez 
pris connaissance de la licence CeCILL, et que vous en avez accept les
termes.

lightness1024@free.fr
--------------------------------------------------------
*/

#ifndef _SECUREARRAY_H_
#define _SECUREARRAY_H_

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <new>

#ifndef NULL
#define NULL 0
#endif

class OutOfRangeException;

//----------------------------------------------------
// Name: SecureArray
// type: template Class
// desc: gestion scurise des tableaux
//----------------------------------------------------
template <typename Any>
class SecureArray
{
private:
	Any* data;
	int elements;
	int uBound;
	int reserved;

	int TotalMem(void) const
	{  // inline
		return elements + reserved;
	}
	void Release(void);
	void RebuildExtra(int bound = -1);
	void DestructExtra(int bound = -1);

public:
	SecureArray(void);
	SecureArray(int NumberOfElements);
	SecureArray(int NumberOfElements, const Any& value);
	SecureArray(const SecureArray<Any>& SArr);
	~SecureArray(void);
	Any& operator [] (int Index);
	const Any& operator [] (int Index) const;
	Any& Back(void);
	const Any& Back(void) const;
	SecureArray<Any>& operator = (const SecureArray<Any>& Source);
	void Resize(int NumberOfElements, bool preserve = true);
	Any* GetDataAddress(int Index = 0) const;
	Any& GetElement(int Index) const;
	const Any& SetElement(int Index, const Any& value);
	int GetUbound(void) const;
	int GetElementsNumber(void) const;
	void Initialize(const Any& value, int Start = 0, int End = -1);
	void Copy(const SecureArray<Any>& Source, int SourceStart = 0, int DestStart = 0, int ElementsNumber = -1);
	const Any& Insert(int Index, const Any& value);
	const Any& Insert(int Index);
	void Remove(int Index, int num = 1);
	const Any& Add(const Any& value);
	void Add(void);
	void ReserveExtra(int size);
	void FreeExtra(void);
};

// -- mthodes prives --

template <typename Any>
void SecureArray<Any>::RebuildExtra(int bound)
{	// reconstruction d'objets sur le bloc
	// bound est la limite  partir de elements
	if (bound <= -1 || bound > reserved)
		bound = reserved;	
	for (int i = 0; i < bound; ++i)
	{
		new (data + elements + i) Any;  // placement new
	}
}

template <typename Any>
void SecureArray<Any>::DestructExtra(int bound)
{	// destruction des objets en surplus sur le bloc
	// bound est la limite  partir de elements
	if (bound <= -1 || bound > reserved)
		bound = reserved;
	for (int i = bound - 1; i >= 0; --i)
	{
		data[elements + i].~Any();  // destruction explicite /!\ danger ne pas deleter[] brutalement le bloc
	}
}

template <typename Any>
void SecureArray<Any>::Release(void) // prive
{	// dtruit toutes les donnes contenues
	// et dsalloue le pool mmoire.

	// si une partie du bloc est en "terrain vague", il faut la reconstruire avant de dtruire le bloc
	RebuildExtra();
	delete [] data;  // destruction des objets, et dsallocation du bloc
	data = NULL;
	elements = 0;
	reserved = 0;
	uBound = 0;
}

// -- mthodes publiques --

template <typename Any>
SecureArray<Any>::SecureArray(void)
{
	elements = 0;
	uBound = 0;
	reserved = 0;
	data = NULL;
}

template <typename Any>
SecureArray<Any>::SecureArray(int NumberOfElements)
{
	reserved = 0;
	if (NumberOfElements <= 0)
	{
		elements = 0;
		uBound = 0;
		data = NULL;
		return;
	}
	data = new Any[NumberOfElements];
	elements = NumberOfElements;
	uBound = elements - 1;
}

template <typename Any>
SecureArray<Any>::SecureArray(int NumberOfElements, const Any& value)
{
	reserved = 0;
	if (NumberOfElements <= 0)
	{
		elements = 0;
		uBound = 0;
		data = NULL;
		return;
	}
	data = new Any[NumberOfElements];
	elements = NumberOfElements;
	uBound = elements - 1;
	Initialize(value);
}


template <typename Any>
SecureArray<Any>::SecureArray(const SecureArray<Any>& SArr)
{	// constructeur de copie
	elements = SArr.elements;
	uBound = SArr.uBound;
	reserved = 0;  // la copie fait perdre l'extra
	data = NULL;
	if (elements == 0) return;
	data = new Any[elements];
	for (int i = 0; i < elements; i++)
	{
		data[i] = SArr.data[i];
	}
}

template <typename Any>
SecureArray<Any>::~SecureArray(void)
{
	Release();
}

template <typename Any>
void SecureArray<Any>::Resize(int NumberOfElements, bool preserve)
{
	int i;
	Any* NewData = NULL;

	if (NumberOfElements <= 0)
	{  // cas particulier : une redim  zro implique une demande de dsallocation totale
		Release();
	}
	else if (preserve && NumberOfElements <= TotalMem())
	{  // on garde le mme bloc mmoire
		if (NumberOfElements > elements)
		{
			// on regagne du terrain par rapport  l'ancien tableau il faut reconstruire les objets
			RebuildExtra(NumberOfElements - elements);
			reserved = TotalMem() - NumberOfElements;
			elements = NumberOfElements;
			uBound = elements - 1;
		}
		else if (NumberOfElements < elements)
		{
			// les objets en surplus sont dtruits, laissant une partie du bloc en "terrain vague"
			int bound = elements - NumberOfElements;  // sauvegarde temporaire de la borne
			reserved = TotalMem() - NumberOfElements;
			elements = NumberOfElements;
			uBound = elements - 1;
			DestructExtra(bound); // destruction explicite des objets en trop
		}
	}
	else
	{
		if (!preserve)
		{  // suppression avance, pour ne pas occuper la mmoire avec deux blocs simultanment
			Release();
		}
		//  ce stade, NumberOfElements est forcment non nul positif:
		NewData = new Any[NumberOfElements];
		if (preserve)
		{  // rcupration des anciennes donnes
			// le nouveau bloc est forcment plus gros que l'ancien sinon il n'y aurait pas rallocation du bloc
			for (i = 0; i < elements; i++)
			{
				NewData[i] = data[i];
			}
			Release();  // on supprime l'ancien tableau aprs la copie
		}
		elements = NumberOfElements;
		uBound = elements - 1;
		reserved = 0;  // plus de rserve car le bloc n'tait pas preserv ou plus gros que l'ancien
		data = NewData;
	}
}

template <typename Any>
Any* SecureArray<Any>::GetDataAddress(int Index) const
{
	if (Index <= uBound && elements != 0 && Index >= 0)
	{
		return &data[Index];
	}
	else
	{
		return (Any*)NULL;
	}
}

template <typename Any>
Any& SecureArray<Any>::GetElement(int Index) const
{
	if (Index <= uBound && elements != 0 && Index >= 0)
	{
		return data[Index];
	}
	else
	{
		throw OutOfRangeException(Index, elements, "Le programme tente de lire l'index %d alors que le tableau a une taille de %d.");
	}
}

// mutateur:
template <typename Any>
const Any& SecureArray<Any>::SetElement(int Index, const Any& value)
{
	if (Index <= uBound && elements != 0 && Index >= 0)
	{
		data[Index] = value;
		return data[Index];
	}
	else
	{
		throw OutOfRangeException(Index, elements, "Le programme tente d'assigner une valeur  l'index %d alors que le tableau a une taille de %d.");
	}
}

// accesseur:
template <typename Any>
int SecureArray<Any>::GetUbound(void) const
{
	return uBound;
}

template <typename Any>
int SecureArray<Any>::GetElementsNumber(void) const
{
	return elements;
}

template <typename Any>
Any& SecureArray<Any>::operator [] (int Index)
{
	if (Index <= uBound && elements != 0 && Index >= 0)
	{
		return data[Index];
	}
	else
	{
		throw OutOfRangeException(Index, elements, "Le programme tente d'assigner une valeur  l'index %d alors que le tableau a une taille de %d.");
	}
}

template <typename Any>
const Any& SecureArray<Any>::operator [] (int Index) const
{
	if (Index <= uBound && elements != 0 && Index >= 0)
	{
		return data[Index];
	}
	else
	{
		throw OutOfRangeException(Index, elements, "Le programme tente de lire l'index %d alors que le tableau a une taille de %d.");
	}
}

template <typename Any>
Any& SecureArray<Any>::Back(void)
{
	if (elements != 0)
		return data[uBound];
	else
		throw OutOfRangeException(uBound, elements, "Le programme tente de lire l'index %d alors que le tableau a une taille de %d.");
}

template <typename Any>
const Any& SecureArray<Any>::Back(void) const
{
	if (elements != 0)
		return data[uBound];
	else
		throw OutOfRangeException(uBound, elements, "Le programme tente de lire l'index %d alors que le tableau a une taille de %d.");
}


template <typename Any>
void SecureArray<Any>::Initialize(const Any& value, int Start, int End)
{
	if (Start < 0)
	{
		Start = 0;
	}
	if (End == -1)
	{
		End = uBound;
	}
	if (Start > uBound || End > uBound)
	{
		int val = (Start > End) ? Start : End;
		throw OutOfRangeException(val, elements, "Le programme tente d'initialiser jusqu'a l'index %d alors que le tableau a une taille de %d.");
	}
	for (int i = Start; i <= End; i++)
	{
		data[i] = value;
	}
}

template <typename Any>
void SecureArray<Any>::Copy(const SecureArray<Any>& Source, int SourceStart, int DestStart, int ElementsNumber)
{
	if (ElementsNumber == - 1)
	{
		ElementsNumber = Source.elements - SourceStart;
		if (ElementsNumber < 0)
			ElementsNumber = 0;
	}
	if (SourceStart < 0)
	{
		throw OutOfRangeException(SourceStart, 0, "Copie de bloc a partir de %d alors qu'on ne peut partir que de %d au moins.");
	}
	if (SourceStart + ElementsNumber > Source.elements)
	{
		throw OutOfRangeException(SourceStart + ElementsNumber, Source.elements, "Copie de bloc jusqu'a %d alors que le tableau a une taille de %d.");
	}
	if (DestStart + ElementsNumber > elements)
	{
		Resize(DestStart + ElementsNumber, true);
	}
	if (DestStart < 0)
	{
		ElementsNumber += DestStart;
		DestStart = 0;
	}
	for (int i = 0; i < ElementsNumber; ++i)
	{
		data[DestStart + i] = Source.data[SourceStart + i];
	}
}

template <typename Any>
SecureArray<Any>& SecureArray<Any>::operator = (const SecureArray<Any>& Source)
{
	Release();
	Copy(Source, 0, 0, Source.elements);
	return *this;
}

template <typename Any>
const Any& SecureArray<Any>::Insert(int Index, const Any& value)
{
	if (Index < 0 || Index > uBound || elements == 0)
	{
		throw OutOfRangeException(Index, elements, "Le programme tente d'inserer une valeur  l'index %d alors que la taille est de %d.");
	}
	if (reserved <= 0)
	{
		ReserveExtra(elements >> 1);
	}
	Resize(elements + 1, true);
	for (int i = uBound; i > Index; i--)
	{	// on remonte toutes les infos:
		data[i] = data[i - 1];
	}
	// et on pose l'element:
	data[Index] = value;
	return data[Index];
}

template <typename Any>
const Any& SecureArray<Any>::Insert(int Index)
{
	Any defaultElem;
	return Insert(Index, defaultElem);
}

template <typename Any>
void SecureArray<Any>::Remove(int Index, int num)
{
	if (num <= 0)
		return;
	if (Index < 0 || Index + num > elements || elements == 0)
	{
		throw OutOfRangeException(Index + num, elements, "Le programme tente de retirer une valeur  l'index %d alors que la taille du tableau est %d.");
	}
	for (int i = Index; i < elements - num; i++)
	{	// on redescend toutes les infos:
		data[i] = data[i + num];
	}
	Resize(elements - num, true);
}

template <typename Any>
const Any& SecureArray<Any>::Add(const Any& value)
{
	if (reserved <= 0)
	{
		ReserveExtra(elements >> 1);
	}
	Resize(elements + 1, true);
	data[uBound] = value;
	return data[uBound];
}

template <typename Any>
void SecureArray<Any>::Add(void)
{
	if (reserved <= 0)
	{
		ReserveExtra(elements >> 1);
	}
	Resize(elements + 1, true);
}

template <typename Any>
void SecureArray<Any>::ReserveExtra(int size)
{
	int oldE, oldB, oldR;
	
	if (size > 0)
	{
		oldE = elements;
		oldR = reserved;
		oldB = uBound;
		Resize(TotalMem() + size, true);
		reserved = elements - oldE;
		elements = oldE;
		uBound = oldB;
		DestructExtra();
	}
}

template <typename Any>
void SecureArray<Any>::FreeExtra(void)
{
	int i;
	int oldE;
	Any* ptr;

	if (reserved >= 0)
	{
		if (elements <= 0)
		{
			Release();
		}
		else
		{
			oldE = elements;
			ptr = new Any[elements];
			for (i = 0; i < elements; ++i)
			{
				ptr[i] = data[i];
			}
			Release();
			elements = oldE;
			uBound = oldE - 1;
			reserved = 0;
			data = ptr;
		}
	}
}

//----------------------------------------------------
// name: OutOfRangeException
// type: Class
// desc: gestion des exceptions de la classe SecureArray
//----------------------------------------------------
class OutOfRangeException
{
protected:
	int BadIndex;
	int Elements;
	const char* ErrorString;
public:
	OutOfRangeException(int BadIndex, int Elements, const char* ErrorString = NULL);
	int GetBadIndex(void);
	int GetElements(void);
	void GetErrorStringPtr(char* buffer, int BuffSize);
};

#endif
