/**********************************************************************
 * File  : carsmanager.cpp
 * Author: Vivien - a.k.a. Lightness1024!
 *         mailto:Lightness1024@free.fr
 * Date  : 12/07/2007
 *********************************************************************/

#include "carsmanager.hpp"

// ------------------------------------------------
// class "Car"
// ------------------------------------------------

// il revient a l'utilisateur d'updater les voitures
// qu'il a instancie.
void Car::Update(float delta_time, TerrainManager* tmgr)
{
	Car copy(*this);  // on copie l'etat actuel au cas ou le mouvement est refuse par une collision
	Vector2f dirbackup;
	float anglebackup;

	// direction :
	angle += GetWheelFactor() * delta_time * speed;     // radians/s
	direction = Rotate(Vector2f(0.f, -1.f), angle);
	// position :
	posf += direction * speed * 50 * delta_time;  // calcul necessairement floating point
	// cast en pixels :
	rect.pos.x = (int)posf.x;
	rect.pos.y = (int)posf.y;
	
	// calcul du ORect:
	ORect orc(Rect(0, 0, 40, 24));
	orc.rect.SetCenter(rect.GetCenter());
	orc.dir = direction;
	
	dirbackup = direction;
	anglebackup = angle;
	// si collision, rollback dans l'etat precedent la tentative de mouvement :
	if (tmgr->WorldIntersectORect(orc))
	{
		// probleme : si on a pas le droit de tourner en collision, on est coince pour toujours.
		// workaround : si l'user veut tourner, testons si les positions decalees d'un pixel
		//              en arriere du sens de marche sont libres.
		
		if (copy.angle != anglebackup)  // si on veut tourner
		{
			int ddx[4] = {0, 1, -1, -1};
			int ddy[4] = {-1, 1, 1, -1};
			int dx[4] = {0, 1, 0, -1};
			int dy[4] = {-1, 0, 1, 0};
			bool free = false;
			for (int i = 0; i < 4; ++i)  // tester 4 deplacements
			{
				orc.rect.pos.x += ddx[i];
				orc.rect.pos.y += ddy[i];

				if (!tmgr->WorldIntersectORect(orc))  // liiiibre !
				{
					posf.x += dx[i];
					posf.y += dy[i];
					rect.pos.x = (int)posf.x;
					rect.pos.y = (int)posf.y;
					free = true;
					break;
				}
			}
			if (!free)
			{ // bon ben solution de secours:
				*this = copy;
				angle = anglebackup;
				direction = dirbackup;
			}
		}
		else
			*this = copy;  // blocage
	}
}

// ------------------------------------------------
// class "CarsManager"
// ------------------------------------------------

void CarsManager::DrawCars(void)
{
	for (CarIT it = cars.begin(); it != cars.end(); ++it)
	{
		// positionner les sprites :
		sprites.Rotate(it->spriteId, -ToDegree(it->car.angle));
		sprites.Move(it->spriteId, it->car.rect.pos.x, it->car.rect.pos.y);
		// et dessiner :
		sprites.Draw(it->spriteId);
	}
}

// si l'utilisateur ne se sert QUE d'un CarsManager dans tout le programme
// (et donc d'aucun autre sprite manager) il peut vouloir rafraichir
// la surface SDL, donc il faut permettre un access a Update.
void CarsManager::Update(UPDATE_PARAM_en p)
{
	sprites.Update(p);  // oui, car Update - time only ne sert a rien du tout pour les sprites rotates.
}

CarsManager::CarAcsor CarsManager::Push(const char* sprite)
{
   cars.push_front(ManagedCar());  // on pousse en tete pour pouvoir acceder a l'iterateur en temps constant
   cars.front().spriteId = sprites.LoadSprite(sprite, SP_ROTATED);  // les sprites de voitures sont de type "rotated"
   // renseignement du rect :
   cars.front().car.rect.w = sprites.GetW(cars.back().spriteId);
   cars.front().car.rect.h = sprites.GetH(cars.back().spriteId);
   // on retourne un accesseur :
   return CarAcsor(cars.begin());  // construit avec l'iterateur, qui ne sera jamais invalide. (voir doc SGI)
}
// suppression d'une voiture dans la liste:
void CarsManager::Pop(CarAcsor& acsor)
{
   if (acsor.valid)
   {
      cars.erase(acsor.it);  // temps constant, n'invalide pas les autres iterateurs.
      acsor.valid = false;
   }
}
