/**********************************************************************
 * File  : carsmanager.hpp
 * Author: Vivien - a.k.a. Lightness1024!
 *         mailto:lightness1024@free.fr
 * Date  : 12/07/2007
 *********************************************************************/

#ifndef CARS_MANAGER_NOVEND_
#define CARS_MANAGER_NOVEND_ 1

#include <list>
#include "geometry.hpp"
#include "spritemanager.hpp"
#include "terraingenerator.hpp"
#include "timer.hpp"

using std::list;


// ca c'est publique
// public intel

struct Car
{
		enum DIRECTION_en
		{
			DIR_STRAIGHT,
			DIR_RIGHT,
			DIR_LEFT
		};
		
	private:
		
		float GetWheelFactor(void)
		{
			switch (wheel)
			{
				case DIR_RIGHT:
					return -1.0f;
				case DIR_LEFT:
					return +1.0f;
				default:
					return 0.0f;
			}
		}
		
		DIRECTION_en wheel;
		Vector2f posf;
		
	public:

		Rect rect;
		Vector2f direction;
		float angle;
		float speed;
		
		void Update(float delta_time, TerrainManager* tmgr);
		void SetDirection(DIRECTION_en edir)
		{
			wheel = edir;
		}
		void SetPosition(int x, int y)
		{
			posf.x = x;
			posf.y = y;
			rect.pos.x = x;
			rect.pos.y = y;
		}

		Car(void)
				: wheel(DIR_STRAIGHT),
				  posf(0, 0),
				  angle(0.f),
				  speed(4.5f) { }
};


// ---------------------------

class CarsManager
{
	protected:

		struct ManagedCar
		{
				Car car;
				int spriteId;
		};

		SpriteManager sprites;
		typedef list< ManagedCar >::iterator CarIT;
		list< ManagedCar > cars;
		
		CTimer time;
		float prev_time;
		
	public:
		
		// structure a renvoyer a l'utilisateur, il peut acceder aux voitures et demander une suppression
		// en temps constant grace a ca.
		struct CarAcsor  // Car Accessor.
		{
				friend class CarsManager;

				Car& operator * (void)
				{
					return it->car;
				}
				Car* operator -> (void)
				{
					if (valid)
						return &it->car;
					else
						return NULL;
				}
				
				// constructeurs publiques.
				CarAcsor(const CarAcsor& op2)
						: it(op2.it),
						  valid(op2.valid) { }
				
				// la construction par defaut ne donne pas d'accesseur valide.
				CarAcsor(void)
				{
					valid = false;
				}
				CarAcsor operator = (const CarAcsor& op2)
				{
					it = op2.it;
					valid = op2.valid;
					return *this;
				}

			private:
				CarIT it;  // iterateur de std::list.
				bool valid;

				CarAcsor(CarIT a_it)  // prive, mais CarManager peut en instancier
						: it(a_it),
						  valid(true) { }   // seul ce constructeur prive peut valider un accesseur.
		};

		CarAcsor Push(const char* sprite);  // file name pour le sprite
		// suppression d'une voiture dans la liste:
		void Pop(CarAcsor& acsor);
      
      void DrawCars(void);
		void Update(UPDATE_PARAM_en p = UP_TIME_ONLY);

		CarsManager(void)
		{
			time.Start();
			prev_time = time.Read();
		}
};

#endif
