/**********************************************************************
 * File  : hostilesmanager.cpp
 * Author: Vivien - a.k.a. Lightness1024!
 *         mailto:lightness1024@free.fr
 * Date  : 12/07/2007
 *********************************************************************/

#include "hostilesmanager.hpp"

#include <iostream>
using namespace std;

extern CarsManager::CarAcsor g_user;


bool HostilesManager::SpawnHostile(SPAWN_POSITION_en spawnPos, int isprite)
{
	// creation d'une voiture :
	CarsManager::CarAcsor newcar = rfCarsMngr.Push(spriteFiles[isprite]);
   hostiles.push_front(newcar);
	const int bordw = rfTerrMngr.GetBorderWidth();
	const Vector2i terrSize = rfTerrMngr.GetSize();
	
	// faire spawner les ennemis a la toute droite du niveau
	if (spawnPos == SP_LEVEL_RIGHT)
	{
		// position aleatoire en vertical, mais 500 pixel a droite du bout du niveau
		newcar->SetPosition(terrSize.x + 500,
								  bordw +
								  rnd.Get(terrSize.y - bordw * 2));
	}
	else // MIDDLE
	{
		Rect rc(newcar->rect);
		
		// meme technique que dans le generateur de terrain
		int security = 0;
		bool intersect = true;
		const int max_repeats = terrSize.x / rc.w + terrSize.y / rc.h;  // heuristique rapide
		while (intersect && security < max_repeats)  // jusqu'a que ca rentre
		{
			// choix aleatoire : position sur la carte
			rc.pos.x = bordw + rnd.Get(terrSize.x - rc.w - (bordw << 1));
			rc.pos.y = bordw + rnd.Get(terrSize.y - rc.h - (bordw << 1));
			intersect = rfTerrMngr.IsOverObject(rc);
			++security;
		}
		if (security == max_repeats)  // pas marche
		{
         hostiles.pop_front();
			return false;
		}
      newcar->SetPosition(rc.pos.x,
								  rc.pos.y);
	}
   // reglage de la cible :
   hostiles.front().target.x = 0;
   hostiles.front().target.y = terrSize.y / 2;
	// angle aleatoire :
	newcar->angle = rnd.Get(359);
   
	return true;
}

void HostilesManager::SetSpawnRule(int spawnRules_flag, int flagParameter)
{
   options = spawnRules_flag;
   if (options & SPR_FREQUENCY)
   {
      frequency = flagParameter;
   }
}

void HostilesManager::Update(float delta_time)
{
   timeAccumulator += delta_time;
   if (!(options & SPR_DISABLED))
   {
      // si on depasse le temps prevu pour le prochain spawn, alors on spawn :
      if (timeAccumulator >= nextSpawn && spriteFiles.size() > 0)
      {
         // determiner le temps du prochain spawn :
			if (frequency == 0)
				nextSpawn = 1e37f;
			else
				nextSpawn = rnd.Get(12000 / frequency) * 0.01f;  // frequency est un nombre d'ennemi par minute moyen.
         timeAccumulator = 0.f;  // reset
         
         bool mid = options & SPR_MIDDLE;
         bool rig = options & SPR_RIGHT;
         if (mid && rig) // il faut decider lequel
         {
            if (rnd.Get(1) == 1)
               rig = false;
            else
               mid = false;
         }
         if (rig)  // a droite
         {
            SpawnHostile(SP_LEVEL_RIGHT, rnd.Get(spriteFiles.size() - 1));
         }
         if (mid)  // au milieu
         {
            SpawnHostile(SP_LEVEL_MIDDLE, rnd.Get(spriteFiles.size() - 1));
         }
      }
   }

   // calculer les mouvements de chaque hostile
   list< Hostile >::iterator it;
   for (it = hostiles.begin(); it != hostiles.end(); ++it)
   {
      IntelligenceAndReaction(*it);
      it->pcar->Update(delta_time, &rfTerrMngr);
      
      // regle de disparition : arriver pres du bord
      if (it->pcar->rect.pos.x <= rfTerrMngr.GetBorderWidth() + 10)
      {
         rfCarsMngr.Pop(it->pcar);
         it = hostiles.erase(it);
         --it;
      }
   }
}

void HostilesManager::IntelligenceAndReaction(Hostile& hostile)
{
	if (!hostile.CheckOverride(now))
	{
		// intelligence de base : se diriger vers la cible
		Vector2i toTarget(hostile.target - hostile.pcar->rect.GetCenter());
		Vector2i ortho(toTarget.GetOrtho());
		// on projette la direction actuelle de la voiture sur le vecteur qui va vers la cible
		// puis on va a gauche si positif, et a droite si negatif.
		Vector2i intCastedDirection(hostile.pcar->direction * 15);  // * 15 sinon toujours 0 vu que direction est unitaire
		int proj = intCastedDirection * ortho;  // produit scalaire
		int reverse = intCastedDirection * toTarget;  // si positif, bonne direction, sinon, reverse.
		if (proj > 0 || proj == 0 && reverse < 0)
		{
			hostile.pcar->SetDirection(Car::DIR_LEFT);
		}
		else if (proj < 0)
		{
			hostile.pcar->SetDirection(Car::DIR_RIGHT);
		}
		else
		{
			hostile.pcar->SetDirection(Car::DIR_STRAIGHT);
		}
	}

	// evitement d'obstacle :
	// lancer trois rayons devant, prendre le plus petit en dessous d'un certain seuil
	// tourner du cote opose
	Vector2f start1, start2, start3;
	Vector2f v;
	Vector2f dir = hostile.pcar->direction;
	start1 = hostile.pcar->rect.GetCenter();
	v = dir * 5;
	start1 -= v;
	start3 = start2 = start1;
	start1 += dir.GetOrtho() * 12;  // a peu pres les "phares" de la voiture
	start2 -= dir.GetOrtho() * 12;
	float dist1 = 1e37f;
	float dist2 = 1e37f;
	float dist3 = 1e37f;
	rfTerrMngr.WorldIntersectRay(&dist1, start1, dir);
	rfTerrMngr.WorldIntersectRay(&dist2, start2, dir);
	rfTerrMngr.WorldIntersectRay(&dist3, start3, dir);
	
	if (dist3 < 30) // probablement coince
	{
		hostile.pcar->SetDirection(Car::DIR_LEFT);
		hostile.ActiveOverride(now, 2.f);  // demi tour
	}
	else
	{
		if (dist1 < dist2)
		{
			if (dist1 < 80)
			{
				hostile.pcar->SetDirection(Car::DIR_LEFT);
				hostile.ActiveOverride(now, 0.8f);
			}
		}
		else
		{
			if (dist2 < 80)
			{
				hostile.pcar->SetDirection(Car::DIR_RIGHT);
				hostile.ActiveOverride(now, 0.8f);
			}
		}
	}
	hostile.target = g_user->rect.pos;
}

void Hostile::ActiveOverride(CTimer& timeref, float duration)
{
	if (!overrideOn)
	{
		overrideOn = true;
		stamp = timeref.Read();
		overrideDuration = duration;
	}
}

bool Hostile::CheckOverride(CTimer& timeref)
{
	if (overrideOn)
	{
		if (stamp + overrideDuration < timeref.Read())
		{
			overrideOn = false;
		}
	}
	return overrideOn;
}
