/**********************************************************************
 * File  : hostilesmanager.hpp
 * Author: Vivien - a.k.a. Lightness1024!
 *         mailto:lightness1024@free.fr
 * Date  : 12/07/2007
 *********************************************************************/

#ifndef HOSTILES_NOV_HPP_
#define HOSTILES_NOV_HPP_ 1

#include "carsmanager.hpp"
#include "terraingenerator.hpp"
#include "randclass.hpp"

#include <vector>
using std::vector;
#include <list>
using std::list;

enum SPAWN_POSITION_en
{
	SP_LEVEL_RIGHT,
	SP_LEVEL_MIDDLE
};

enum SPAWN_RULES_en
{
   // option flags
   SPR_DISABLED  = 1,
   SPR_RIGHT     = 2,
   SPR_MIDDLE    = 4,
   SPR_BOTH      = 6,
   SPR_FREQUENCY = 8  // hostiles par minute
};

struct Hostile
{
      CarsManager::CarAcsor pcar;
      Vector2i target;
		
		void ActiveOverride(CTimer& timeref, float duration);
		bool CheckOverride(CTimer& timeref);

      Hostile(void){}
      Hostile(const CarsManager::CarAcsor& a_pcar)
            : pcar(a_pcar),
				  overrideOn(false) { }
		
	private:

		float stamp;
		bool overrideOn;
		float overrideDuration;
};

class HostilesManager
{
		CarsManager& rfCarsMngr;
		TerrainManager& rfTerrMngr;
      
      list< Hostile > hostiles;
		
		CRandom rnd;
      int options;
      int frequency;
      float timeAccumulator;
      float nextSpawn;
		
		CTimer now;

	public:

		vector< const char* > spriteFiles;
		
		// retourne faux si il n'a pas ete possible d'inserer un ennemi dans le niveau
		bool SpawnHostile(SPAWN_POSITION_en spawnPos, int indexSprite);
      
      void SetSpawnRule(int spawnRules_flag, int flagParameter = 0);

      void Update(float delta_time);
      
      void IntelligenceAndReaction(Hostile& hostile);
		
		HostilesManager(CarsManager& a_rCarsMngr,
							 TerrainManager& a_rTerrMngr)
				: rfCarsMngr(a_rCarsMngr),
				  rfTerrMngr(a_rTerrMngr),
              timeAccumulator(0.f),
              nextSpawn(1.f)
      {
         SetSpawnRule(SPR_BOTH | SPR_FREQUENCY, 6);
			now.Start();
      }
};


#endif
