/**********************************************************************
 * File  : solardiales.cpp
 * Author: Vivien - a.k.a. Lightness1024!
 *         mailto:lightness1024@free.fr
 * Date  : 09/07/2007
 *********************************************************************/

#include "spritemanager.hpp"
#include <iostream>
#include "randclass.hpp"
#include "terraingenerator.hpp"
#include "carsmanager.hpp"
#include "hostilesmanager.hpp"
#include "geometry.hpp"
#include <string.h>

using std::cout;
using std::endl;

#define RES_X 1270
#define RES_Y 950

// ------------------------
// handles screen fancy functions
// like soft scroll pusing
// ------------------------
class Screen
{
		Rect src;

		Vector2f pos;
		int repulsion_d;
		float mass;		
		Vector2f speed;
		Vector2f force;

	public:
		
		void SetRepulsionDistance(int d)
		{
			repulsion_d = d;
		}
		
		void MoveMagnetCenter(const Vector2i& center);
		const Rect& GetRc(void) const
		{
			return src;
		}

		void Update(float delta_time);
		
		// on donne les infos a la construction :
		Screen(int resX, int resY, int initCenterX, int initCenterY)
				: src(initCenterX - resX / 2,
						initCenterY - resY / 2,
						resX,
						resY),
				  pos(Vector2f(src.pos.x, src.pos.y)),  // int a float
				  repulsion_d(Min(resX, resY) / 2),
				  mass(0.1f),
				  speed(0, 0)
		{ }
};

void Screen::MoveMagnetCenter(const Vector2i& center)
{
	const int demirx = src.w >> 1;  // div2
	const int demiry = src.h >> 1;
	Vector2i inRect = center - src.pos;
	// recuperation des distances vers les paroies les plus proches :
	int vx, vy;
	Vector2f oposite(1.f, 1.f);
	if (inRect.x < demirx)
	{
		vx = inRect.x;
		oposite.x = -1.f;
	}
	else
	{
		vx = src.w - inRect.x;
	}
	if (inRect.y < demiry)
	{
		vy = inRect.y;
		oposite.y = -1.f;
	}
	else
	{
		vy = src.h - inRect.y;
	}

	// determiner le vecteur force :
	force.x = (float) repulsion_d - Min(vx, repulsion_d);
	force.y = (float) repulsion_d - Min(vy, repulsion_d);
	force.x *= oposite.x;
	force.y *= oposite.y;
}

void Screen::Update(float delta_time)
{
	Vector2f accel;
   Vector2f friction;
   Vector2f result;
   
   friction = speed * 0.7;
   result = force - friction;
	accel = result * (1.f / mass);
	speed += accel * delta_time;
	pos += speed * delta_time;
	src.pos.x = (int)pos.x;
	src.pos.y = (int)pos.y;
}

int gameover;
Vector2<float> g_scroll(0, 0);
Vector2<float> g_scrPos(0, 0);
CarsManager::CarAcsor g_user;
bool g_cheat = false;;

void HandleEvent(SDL_Event event)
{
	switch (event.type)
	{
		/* close button clicked */
		case SDL_QUIT:
			gameover = 1;
			break;

			/* handle the keyboard */
		case SDL_KEYDOWN:
			switch (event.key.keysym.sym)
			{
				case SDLK_ESCAPE:
				case SDLK_q:
					gameover = 1;
					break;
				case SDLK_LEFT:
					g_user->SetDirection(Car::DIR_LEFT);
					break;
				case SDLK_RIGHT:
					g_user->SetDirection(Car::DIR_RIGHT);
					break;
				case SDLK_UP:
					if (g_cheat)
						g_user->speed += 1.5;
					break;
				case SDLK_DOWN:
					if (g_cheat)
						g_user->speed -= 1.5;
					break;
				default:
					break;
			}
			break;
		case SDL_KEYUP:
			switch (event.key.keysym.sym)
			{
				case SDLK_LEFT:
					g_user->SetDirection(Car::DIR_STRAIGHT);
					break;
				case SDLK_RIGHT:
					g_user->SetDirection(Car::DIR_STRAIGHT);
					break;
					/*case SDLK_UP:
					  g_scroll.y = 0;
					  break;
					  case SDLK_DOWN:
					  g_scroll.y = 0;
					  break;*/
				default:
					break;
			}
			break;
		default:
			break;
	}
}


int main(int argc, char* argv[])
{
	SpriteManager sprmngr(RES_X, RES_Y, "solardiales");
	Screen screen(SpriteManager::GetResX(), SpriteManager::GetResY(), 0, 0);
	CRandom rgen;
	TerrainManager terrain;
	CarsManager carmngr;
	HostilesManager enemies(carmngr, terrain);
   enemies.SetSpawnRule(SPR_FREQUENCY | SPR_MIDDLE, 1300);
	
	enemies.spriteFiles.push_back("../resources/enemy.bmp");
   
	TerrainParameters param;
	param.seed = 0;
	param.size_long = 192;
	param.size_height = 24;

	param.density[KTO_WATER] = 0.03;
	param.density[KTO_TREE] = 0.04;
	param.density[KTO_ROCKS] = 0.02;
	param.density[KTO_BONUS] = 0.001;
	
	param.terrain_sprite = "../resources/terrain.bmp";
	param.wall_sprite = "../resources/wall.bmp";
	
	param.sprites[KTO_WATER].push_back("../resources/water.bmp");
	param.sprites[KTO_WATER].push_back("../resources/water2.bmp");
	
	param.sprites[KTO_TREE].push_back("../resources/tree.bmp");
	param.sprites[KTO_TREE].push_back("../resources/tree2.bmp");
	
	param.sprites[KTO_ROCKS].push_back("../resources/rocks.bmp");
	param.sprites[KTO_ROCKS].push_back("../resources/rocks2.bmp");
	
	param.sprites[KTO_BONUS].push_back("../resources/wrench.bmp");
	param.sprites[KTO_BONUS].push_back("../resources/plasma.bmp");
	
	terrain.Generate(param);
	
	g_user = carmngr.Push("../resources/car.bmp");
	g_user->SetPosition(128, 70);
	
	if (argc == 2)
		if (!strcmp(argv[1], "--cheat"))
			g_cheat = true;
	
	int frames = 0;
	CTimer temps;
	temps.Start();
	
	float last = 0;
	
	while (!gameover)
	{
		SDL_Event event;
		
		// look for an event
		if (SDL_PollEvent(&event)) {
			HandleEvent(event);
		}
		
		g_user->Update(temps.Read() - last, &terrain);
      Vector2i attractCenter(g_user->rect.GetCenter());
      attractCenter.x += (int)(g_user->direction.x * 180.f);
      attractCenter.y += (int)(g_user->direction.y * 180.f);
		screen.MoveMagnetCenter(attractCenter);
		screen.Update(temps.Read() - last);
      enemies.Update(temps.Read() - last);
		last = temps.Read();
		
		SpriteManager::scroll_x = screen.GetRc().pos.x;
		SpriteManager::scroll_y = screen.GetRc().pos.y;
      
		terrain.DrawBackGround();
		carmngr.DrawCars();
		terrain.DrawHighGround();
		terrain.Update();
		carmngr.Update();
		sprmngr.Update();
		++frames;
//		if (frames > 1200)
		//break;
	}

	cout << frames / temps.Read() << " fps" << endl;
	
	return 0;
}


// ****************************************
//             main(s) de test
// ****************************************

/*
int main(void)
{
	Vector2f st1, sg1, st2, sg2;
	
	st1.x = st1.y = 0;
	sg1.x = 10;
	sg1.y = 0;
	st2.x = 5;
	st2.y = -5;
	sg2.x = 0;
	sg2.y = 10;
	
	cout << " collision is " << SegmentsCollides(st1, sg1, st2, sg2);
	return 0;
}
*/

/*
int main(void)
{
	ORect rc(Rect(-1, 4, 10, 10));
	
	rc.dir.x = 0.707;
	rc.dir.y = -0.707;
	
	float t;
	rc.IntersectsRay(&t, Vector2f(5,-5), Vector2f(0, 1));
	cout << " collision distance : " << t << endl;
	return 0;
}
*/

/*
int main(void)
{
	SpriteManager sprmngr(RES_X, RES_Y, "solardiales");
	sprmngr.LoadSprite("../resources/car.bmp", SP_ROTATED);
	sprmngr.LoadSprite("../resources/enemy.bmp", SP_ROTATED);
	SpriteManager sprmngr2;
	sprmngr2.LoadSprite("../resources/tree.bmp");
	sprmngr2.Move(0, 200, 250);

	int frames = 0;
	CTimer temps;
	temps.Start();
	
	float last = 0;
	
	SDL_EnableKeyRepeat(20, 20);
	
	ORect me, notme;
	int posx = 50;
	int posy = 50;
	int angle = 0;
	int notmyangle = 0;
	
	me.rect.w = 24;
	me.rect.h = 40;
	notme.rect = me.rect;
	notme.rect.pos.x = 150;
	notme.rect.pos.y = 100;
	
	while (!gameover)
	{
		SDL_Event event;
		if (SDL_PollEvent(&event))
		{
			switch (event.type)
			{
				case SDL_QUIT:
					gameover = 1;
					break;
				case SDL_KEYDOWN:
					switch (event.key.keysym.sym)
					{
						case SDLK_ESCAPE:
						case SDLK_q:
							gameover = 1;
							break;
						case SDLK_LEFT:
							posx--;
							break;
						case SDLK_RIGHT:
							posx++;
							break;
						case SDLK_UP:
							posy--;
							break;
						case SDLK_DOWN:
							posy++;
							break;
						case SDLK_r:
							angle++;
							break;
						case SDLK_t:
							angle--;
							break;
						case SDLK_f:
							notmyangle++;
							break;
						case SDLK_g:
							notmyangle--;
							break;
						default:
							break;
					}
					break;
			}

		}
		
		sprmngr.ClearScreen();
		
		
		me.rect.SetCenter(Vector2i(posx + 24, posy + 24));
		me.dir = Rotate(Vector2f(1, 0), -ToRadians(angle));
		
		notme.dir = Rotate(Vector2f(1, 0), -ToRadians(notmyangle));
		
		sprmngr.Rotate(0, angle);
		sprmngr.Rotate(1, notmyangle);
		sprmngr.Move(0, posx, posy);
		sprmngr.Move(1, 138, 96);
		sprmngr.DrawAll();
		
		if (notme.IntersectsORc(me))
			sprmngr2.DrawAll();

		last = temps.Read();
		
		sprmngr.Update();
		++frames;
	}

	cout << frames / temps.Read() << " fps" << endl;
	
	return 0;	
}
*/
