/**********************************************************************
 * File  : terraingenerator.cpp
 * Author: Vivien - a.k.a. Lightness1024!
 *         mailto:lightness1024@free.fr
 * Date  : 11/07/2007
 *********************************************************************/

#include "terraingenerator.hpp"

// TODO : enlever
#include <iostream>
using std::cout;
using std::endl;

// procedural generation of terrain
void TerrainManager::Generate(TerrainParameters& params)
{
	if (loaded)
		return;

	// on charge le tile du fond pour decouvrir la taille :
	// we load background tile to discover its size :
	background.LoadSprite(params.terrain_sprite);

	// renseignement des champs de la racine de l'arbre de partitionnement
	// fullfiling of paritionning tree's root fields:
	root.data.rect.pos.x = 0;  // begin high left corner
	root.data.rect.pos.y = 0;
	// ends bottom right corner, so the root span over the full map.
	// le rectangle racine couvre toute la carte.
	levelSize.x = params.size_long * background.GetW(0);
	levelSize.y = params.size_height * background.GetH(0);
	root.data.rect.w = levelSize.x;
	root.data.rect.h = levelSize.y;
	// construction :
	BuildInDepth(root);
	
	// ajout des sprites des murs exterieurs :
	// add external walls :
	
	background.LoadSprite(params.wall_sprite);  // chargement du sprite 1, exceptionellement. pour avoir la taille
	
	const int wallW = background.GetW(1);
	const int wallH = background.GetH(1);
	borderWidth = wallH;
	Rect wrc;
	wrc.w = wallW;
	wrc.h = wallH;
	// on creer 3 paroies
	// we create 3 walls to surround the map.
	for (int x = 1; x < levelSize.x / wallW - 1; ++x)  // horizontal walls
	{
		wrc.pos.x = x * wallW;
		int ref;
		if (x != 1) // cas particulier du sprite charge plus haut
			ref = background.LoadSprite(params.wall_sprite);
		else
			ref = 1;
		wrc.pos.y = 0;
		background.Move(ref, wrc.pos.x, wrc.pos.y);
		InsertObject(new Rocks(wrc, ref));
		
		ref = background.LoadSprite(params.wall_sprite);
		wrc.pos.y = levelSize.y - wallH;
		background.Move(ref, wrc.pos.x, wrc.pos.y);
		InsertObject(new Rocks(wrc, ref));
	}
	for (int y = 0; y < levelSize.y / wallH; ++y)   // pour les murs des cotes. // for side walls
	{
		wrc.pos.y = y * wallH;
		int ref;
		ref = background.LoadSprite(params.wall_sprite);
		wrc.pos.x = 0;
		background.Move(ref, wrc.pos.x, wrc.pos.y);
		InsertObject(new Rocks(wrc, ref));
		// there is no wall on the extreme right.
		// pas de mur a droite, car c'est la ou spawn les ennemis,
		// et c'est aussi en franchissant la ligne qu'on termine.
	}

	// now the random part :
	CRandom rnd(params.seed);  // generateur pseudo aleatoire
	
	// surface counting accumulator for each kind of objects :
	double surface[NB_KIND_OF_TERRAIN_OBJECTS] = {0.};

	const double overall_surface = (double)levelSize.y * levelSize.x; // in pixels
	
	// compute sum of densities for security reasons:
	float sum = 0.f;
	for (int i = 0; i < NB_KIND_OF_TERRAIN_OBJECTS; ++i)
		sum += params.density[i];
	if (sum > 0.5f)
	{
		// there are risks that the las vegas algorithm may not terminate if this sum is too high
		// also, the free surface may become too small for you or the enemies to move on it.
		
		// la generation procedurale tente des positions aux hasard dans l'espace et verifie
		// que les rectangles n'intersectent pas les objets deja poses. si la densite
		// devient trop forte, les essais peuvent s'allonger et l'algorithme devenir lent
		// ou ne plus jamais terminer.
		// de plus, trop de densite reduit la liberte de mouvement dans le jeu.
		params.density[KTO_TREE]  = 0.2;
		params.density[KTO_ROCKS] = 0.15;
		params.density[KTO_WATER] = 0.05;
		params.density[KTO_BONUS] = 0.03;
	}
	
	// bigger objects before, smaller last.
	// d'abord les supposes gros sprites, puis les petits.
	
	// nb de choix parmis les sprites :
	int max_choice;
	CharPtrList::const_iterator it;
	bool intersect;
	// --
	// pas le choix, il faut creer des objets soit avec ce genre de conteneur
	// soit avec un switch.
	// --
	// no choice, either it is this, either it is a switch.
	smart_ptr< ObjInstancier > miniFactory[NB_KIND_OF_TERRAIN_OBJECTS];
	miniFactory[KTO_TREE]  = new Instancier< Tree >;
	miniFactory[KTO_ROCKS] = new Instancier< Rocks >;
	miniFactory[KTO_WATER] = new Instancier< Water >;
	miniFactory[KTO_BONUS] = new Instancier< Bonus >;

	// for all kind of objects
	for (int kind_of_obj = 0; kind_of_obj < NB_KIND_OF_TERRAIN_OBJECTS; ++kind_of_obj)
	{
		// taille de la liste des chaines de noms de fichiers
		max_choice = params.sprites[kind_of_obj].size();

		// boucle de remplissage d'objets sur le terrain :
		while (surface[kind_of_obj] < params.density[kind_of_obj] * overall_surface)
		{
			// premier choix aleatoire : quel sera le sprite representant ?
			// among the list of sprites, let's choose one randomly.
			it = GetRandomIt(params.sprites[kind_of_obj], max_choice, rnd);

			// chargement du sprite :
			// stockage special pour les arbres ; creation d'une reference alias :
			SpriteManager& ground_mngr = (kind_of_obj == KTO_TREE) ? highground : background;
			// indice du sprite dans le spriteManager
			int ref = ground_mngr.LoadSprite(*it);  // ils vont dans les Z eleves
			int large = ground_mngr.GetW(ref);
			int height = ground_mngr.GetH(ref);
			
			Rect newRc;
			intersect = true;
			int security_counter = 0;
			const int max_repeats = (levelSize.x / large) * (levelSize.y / height);  // heuristique
			newRc.w = large;
			newRc.h = height;
			while (intersect && security_counter < max_repeats)  // algo proba type las-vegas
			{
				// deuxieme choix aleatoire : la position sur la carte
				newRc.pos.x = wallW + rnd.Get(levelSize.x - large - (wallW << 1));
				newRc.pos.y = wallW + rnd.Get(levelSize.y - height - (wallH << 1));
				intersect = IsOverObjectFromNode(newRc, root);
				++security_counter;
			}
			if (security_counter == max_repeats)
			{
				break;
			}
			else
			{
				// on sait qu'il est en position libre, ajoutons le au monde:
				InsertObject(miniFactory[kind_of_obj]->Create(newRc, ref));
				// positionnement correct du sprite dans le manager :
				ground_mngr.Move(ref, newRc.pos.x, newRc.pos.y);
				// mise a jour de la surface couverte par les objets :
				surface[kind_of_obj] += large * height;
			}
		}
	}
	
	loaded = true;
}

void TerrainManager::AddFromNode(int index, SectorNode& node)
{
	SectorNode& left = *node.left;
	SectorNode& right = *node.right;
	
	// if object "index" intersects this node
	if (objects[index]->rect.Intersects(node.data.rect))
	{
		if (node.IsLeaf())
		{
			node.data.contents.Add(Sector::FlagIndex(index));  // ajout de l'index a la liste du secteur.
		}
		else
		{
			AddFromNode(index, left);
			AddFromNode(index, right);
		}
	}
}

// check if a rectangle overlaps an object in the world
bool TerrainManager::IsOverObjectFromNode(const Rect& rc, const SectorNode& node)
{
	SectorNode& left = *node.left;
	SectorNode& right = *node.right;

	if (rc.Intersects(node.data.rect))  // if we intersect this node
	{
		if (node.IsLeaf()) // we perform full list-object test only on leaves
		{
			return IsRectIntersectingObjectOfNode(rc, node);
		}
		else
		{
			// maybe we intersect one of the children.
			// if there is something is just one of them, its enough to say true.
			return IsOverObjectFromNode(rc, left)   ||   IsOverObjectFromNode(rc, right);
			// I really hope that lazy OR-expression evaluation is working here.

			// une bonne optimisation ici, est de ne pas evaluer la droite
			// de l'arbre si la gauche a deja retroune "vrai".
			// le compilateur est cense fournir ca.
		}
	}
	else
		return false;  // if we dont even intersect this node, sure we dont intersect any object of it.
}

bool TerrainManager::IsRectIntersectingObjectOfNode(const Rect& rc, const SectorNode& node)
{
	// for all elements in a list of indice contained in a node
	// check the corresponding objects in the world;
	// if one is intersecting, leave and return true.
	int max = node.data.contents.GetElementsNumber();
	for (int i = 0; i < max; ++i)
	{
		if (objects[  node.data.contents[i].index  ]->rect.Intersects(rc))
			return true;
	}
	return false;
}


CharPtrList::const_iterator TerrainManager::GetRandomIt(const CharPtrList& a_list, int max, CRandom& rand) const
{
	CharPtrList::const_iterator it;
	int choice = rand.Get(max - 1);
	it = a_list.begin();
	for (int i = 0; i < choice; ++i)
		++it;  // les iterateurs de liste ne sont pas randoms.
	return it;
}


// construct the tree until leaves reaches reasonable size.
// construction recursive de l'arbre jusqu'a que les feuilles atteignent la bonne taille.
void TerrainManager::BuildInDepth(SectorNode& node)
{
	// we go on if one of the both sides is too long
	// si un des cotes est encore trop grand on continue de couper.
	if (node.data.rect.w > background.GetResX() / 2 || node.data.rect.h > background.GetResX() / 2)
	{
		node.left = new SectorNode;
		node.right = new SectorNode;
		// the longest side is divided by two, to creates two new nodes
		// on decoupe selon le plus grand cote
		if (node.data.rect.w > node.data.rect.h)  // largeur
		{
			node.left->data.rect = node.data.rect;
			node.left->data.rect.w >>= 1;  // div by 2.

			node.right->data.rect = node.left->data.rect;
			node.right->data.rect.pos.x += node.left->data.rect.w;  // begins where "left" ends.
		}
		else  // longueur
		{
			node.left->data.rect = node.data.rect;
			node.left->data.rect.h >>= 1;  // div by 2

			node.right->data.rect = node.left->data.rect;
			node.right->data.rect.pos.y += node.left->data.rect.h;  // same. // pareil.
		}
		// recursion :
		BuildInDepth(*node.left);
		BuildInDepth(*node.right);
	}
}

void TerrainManager::DrawBackGround(void)
{
	// le sprite 0 est le terrain, toujours.
	// sprite 0 si always ground.
	const int tw = background.GetW(0);  // tile w, h
	const int th = background.GetH(0);
	const int offx = (unsigned)background.scroll_x % tw;
	const int offy = (unsigned)background.scroll_y % th;
	for (int x = background.scroll_x; x < background.scroll_x + background.GetResX() + tw; x += tw)
	{
		for (int y = background.scroll_y; y < background.scroll_y + background.GetResY() + th; y += th)
		{
			background.Draw(0, x - offx, y - offy);
		}
	}
	// apres le terrain, les objets:
	// after the ground, objects:
	//   rem : trees are in "highground", drew in a separate function
	//         so that user can draw other things in between the two layers.
	
	// we must determine the list of nodes intersecting the screen.
	Rect rcScreen(background.scroll_x,
					  background.scroll_y,
					  background.GetResX(),
					  background.GetResY());
	
	m_numNodes = 0;
	BuildActivatedNodesList(rcScreen, root, m_nodeList, m_numNodes);
	
	// instead of drawing the full set of objects we have in cache,
	// we just draw the objects contained in the sectors the screen activates.
	
	// on ne dessine que les objets compris dans les noeuds actives par
	// la position de l'ecran.	
	// la recherche en log N permet d'avoir des centaines de milliers de sprites en cache
	// sans pour autant ruiner les performances.
	
	++m_frame;
	for (int node = 0; node < m_numNodes; ++node)
	{
		// j'ai toujours peur que le compilateur ne sache pas optimiser ca :
		const int limit = m_nodeList[node]->data.contents.GetElementsNumber();
		for (int i = 0; i < limit; ++i)
		{
			// alias
			Sector::FlagIndex& fi_ref = m_nodeList[node]->data.contents[i];
			// do not draw 2 times the same sprite on the same frame
			if (fi_ref.flag != m_frame)
			{
				if (!objects[ fi_ref.index ]->highplane)  // not the trees (yet)
				{
					background.Draw(objects[ fi_ref.index ]->spriteId);
					fi_ref.flag = m_frame;
				}
			}
		}
	}      // c'est peut etre pas tres TLB-friendly ce genre de fonctions...
	
	// au lieu de simplement tout dessiner comme ca :
	//background.DrawAll(1);
}

void TerrainManager::DrawHighGround(void)
{
	// il est important que drawhigh soit appelle APRES background
	// car elle utilie la meme liste de nodes actives.
	// --
	// it is important that this function is called AFTER draw backg
	// because it uses the same node list.
	// --
	for (int node = 0; node < m_numNodes; ++node)
	{
		const int limit = m_nodeList[node]->data.contents.GetElementsNumber();
		for (int i = 0; i < limit; ++i)
		{
			Sector::FlagIndex& fi_ref = m_nodeList[node]->data.contents[i];
			if (fi_ref.flag != m_frame)
			{
				if (objects[ fi_ref.index ]->highplane)  // only trees this time
				{
					highground.Draw(objects[ fi_ref.index ]->spriteId);
					fi_ref.flag = m_frame;
				}
			}
		}
	}
	// instead of:
	// 	highground.DrawAll();
}

// log N operation. fill an array of node pointers activated by the overlapping of a rect.
void TerrainManager::BuildActivatedNodesList(const Rect& rc, SectorNode& node, SecureArray< SectorNode* >& nodeList, int& numNodes)
{
	if (rc.Intersects(node.data.rect))
	{
		if (node.IsLeaf())
		{
			if (numNodes > nodeList.GetUbound())  // heuristique de redimensionnement automatique
			{
				nodeList.Resize(nodeList.GetElementsNumber() * 2);  // resize fois 2
			}
			nodeList[numNodes++] = &node;  // copier le pointeur dans la liste
		}
		else
		{
			BuildActivatedNodesList(rc, *node.left, nodeList, numNodes);
			BuildActivatedNodesList(rc, *node.right, nodeList, numNodes);
		}
	}
}

bool TerrainManager::WorldIntersectRect(const Rect& rc)
{
	// comme d'hab :
	// chercher en profondeur les noeuds actives par "rc" et tester les collisions.
	
	m_numNodes = 0;
	BuildActivatedNodesList(rc, root, m_nodeList, m_numNodes);
	
	// donc on y va pour toutes les nodes activees:
	for (int node = 0; node < m_numNodes; ++node)
	{
		const int limit = m_nodeList[node]->data.contents.GetElementsNumber();
		for (int i = 0; i < limit; ++i)  // pour tous les elements dans les listes des nodes
		{
			// alias
			Sector::FlagIndex& fi_ref = m_nodeList[node]->data.contents[i];
			if (objects[ fi_ref.index ]->IntersectRect(rc))
				return true;
		}
	}
	return false;
}

bool TerrainManager::WorldIntersectRay(float* t, const Vector2f& start, const Vector2f& direction, bool lowspan)
{
	// encore comme d'hab...
	// chercher en profondeur les noeuds actives par le rayon et tester les collisions.
	
	// CEPENDANT, on va limiter le rayon a une distance finie d'un tiers de longeur d'ecran
	// autour du debut du rayon.
	
	Rect boundingSpace((int)start.x, (int)start.y, 0 , 0);
	if (lowspan)  // en lowspan on ne couvre que 100 pixels
	{
		boundingSpace.AddPoint(start + direction * 100);
	}
	else
	{
		boundingSpace.AddPoint(start + direction * SpriteManager::GetResY() * 0.7f);
	}
	
	m_numNodes = 0;
	BuildActivatedNodesList(boundingSpace, root, m_nodeList, m_numNodes);
	
	// pour les rayons il faut tout tester
	float mint = 1e37f;
	float argt = 1e37f;
	bool collide = false;
	for (int node = 0; node < m_numNodes; ++node)
	{
		const int limit = m_nodeList[node]->data.contents.GetElementsNumber();
		for (int i = 0; i < limit; ++i)  // pour tous les elements dans les listes des nodes
		{
			Sector::FlagIndex& fi_ref = m_nodeList[node]->data.contents[i];
			if (objects[ fi_ref.index ]->IntersectRay(&argt, start, direction))
			{
				if (argt < mint)
				{
					mint = argt;
				}
				collide = true;
			}
		}
	}
	*t = mint;
	return collide;
}

bool TerrainManager::WorldIntersectORect(const ORect& a_orc)
{
	// et encore encore comme d'hab...
	
	m_numNodes = 0;
	BuildActivatedNodesList(a_orc.GetBoundingRect(), root, m_nodeList, m_numNodes);
	
	for (int node = 0; node < m_numNodes; ++node)
	{
		const int limit = m_nodeList[node]->data.contents.GetElementsNumber();
		for (int i = 0; i < limit; ++i)  // pour tous les elements dans les listes des nodes
		{
			Sector::FlagIndex& fi_ref = m_nodeList[node]->data.contents[i];
			if (objects[ fi_ref.index ]->IntersectORect(a_orc))
				return true;
		}
	}
	return false;
}

